/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northestern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * $Created by: XIAO Tong (xiaotong@mail.neu.edu.cn) 2018-07-17
 */

#include "XBackwardNode.h"
#include "../tensor/core/CHeader.h"
#include "../tensor/XName.h"

namespace nts{

/* make gradient tensor for a node */
void XNoder::MakeGrad(XTensor * node)
{
    if(node == NULL)
        return;

    if(!XTensor::IsIdentical(node, node->grad)){
        delete node->grad;
        node->grad = NewTensor(node);
    }
}

/* the node is a leaf node (intput) or not */
bool XNoder::IsLeaf(XTensor * node)
{
    if(node == NULL)
        return false;

    if(node->income.tailNum == 0)
        return true;
    else
        return false;
}

/* the node is a root node (output) or not */
bool XNoder::IsRoot(XTensor * node)
{
    if(node == NULL)
        return false;

    if(node->outgo.tailNum == 0)
        return true;
    else
        return false;
}

/* the node keeps the gradinent or not */
bool XNoder::IsGrad(XTensor * node)
{
    if(node == NULL)
        return false;
    
    if(node->isGrad)
        return true;
    else
        return false;
}

/* 
compute dE/dx of a node 
Note that input of the node could be arbitrary tensors
>> node - node of the network
*/
void XNodeGrad::Compute(XTensor * node)
{
    if(node == NULL || node->visitMark == NODE_FINISHED)
        return;

    if(!XNoder::IsLeaf(node)){
        

    }

    node->visitMark = NODE_FINISHED;
}

/* indicates whether the node is for a math operation */
bool XNodeGrad::IsMathOP(XTensor * node)
{
    XLink &income = node->income;
    return (income.typeID & MATH_BASE) != 0;
}

/* compute dE/dx as a math operation, e.g., sum, multiply ... */
void XNodeGrad::ComputeMath(XTensor * node)
{
    CheckNTErrors(node->grad != NULL, "No gradient found!");

    XLink &income = node->income;
    int operID = income.typeID;

    /* c = a + b * \beta
       dE/da = dE/dc 
       dE/db = dE/dc * \beta */
    if(operID == MATH_SUM){
        CheckNTErrors(income.tailNum == 2, "Wrong input tensor number for SUM!");
        XTensor * a = income.tails[0]; 
        XTensor * b = income.tails[1];
        DTYPE beta = income.GetParam(0);
        XNoder::MakeGrad(a);
        XNoder::MakeGrad(b);
        _CopyValues(node->grad, a->grad);
        if(beta != 1.0F)
            _ScaleAndShift(node->grad, a->grad, beta);
        else
            _CopyValues(node->grad, b->grad);
    }
    /* c = a * b
       dE/da = dE/dc * b
       dE/db = dE/dc * a */
    else if(operID == MATH_MULTIPLY){
        CheckNTErrors(income.tailNum == 2, "Wrong input tensor number for MULTIPLY!");
        XTensor * a = income.tails[0]; 
        XTensor * b = income.tails[1];
        XNoder::MakeGrad(a);
        XNoder::MakeGrad(b);

        CheckNTErrors(XTensor::IsIdentical(a, b), "Wrong sized input tensors!");
        _Multiply(node->grad, b, a->grad);
        _Multiply(node->grad, a, b->grad);
    }
    else{
        ShowNTErrors("TODO!");
    }
}

}
